from functools import wraps

from sqlalchemy import desc

from ttmr.console import BG_WHITE
from ttmr.console import BLACK
from ttmr.console import GREEN
from ttmr.console import RED
from ttmr.console import Console
from ttmr.db import Database
from ttmr.db import load_incidences
from ttmr.db import regex_factory
from ttmr.ui import _date_input
from ttmr.ui import _option_input
from ttmr.ui import bool_input
from ttmr.ui import date_input
from ttmr.ui import default_text_input
from ttmr.ui import duration_input
from ttmr.ui import number_input
from ttmr.ui import option_input
from ttmr.ui import show_timer
from ttmr.ui import text_input
from ttmr.util import add_table_border

from . import database


def db_session(func):
    @wraps(func)
    def _db_session(cfg):
        Session = database.init(cfg._dbfile)
        cfg._db = Session()

        try:
            return func(cfg)
        finally:
            cfg._db.close()

    return _db_session


def display_config(conf):
    Console.write(str(conf))


def load(conf):
    load_incidences(conf)


def new_incident(conf):
    subject_text = text_input("Email Subject")

    parse_incident_id = regex_factory(
        (
            r"^.*(?P<type>WO|INC|CRQ|FY\d\d)"
            r"0*(?P<number>[0-9]\d+) "
            r".*Description: "
            r"(?P<description>.+)"
            r"(\d+ [KMG]B|\d+:\d\d [AP]M).*$"
        )
    )

    incident_id = parse_incident_id(subject_text) or {}

    Console.newline()
    Console.newline()

    inc_type = default_text_input("Type", incident_id.get("type", "WO"))
    inc_num = number_input("Number", incident_id.get("number", ""))
    inc_desc = default_text_input("Desc", incident_id.get("description", ""))

    Console.newline()

    with Database(conf.sqlite, conf.dt) as db:
        row_update = db.record_incident(inc_type, inc_num, inc_desc)

    if row_update != 1:
        Console.write(RED)
        Console.writeline("Update error. Incicent was not saved.")
        Console.clear_formating()
        return

    Console.write(GREEN)
    Console.write("New incident saved!")
    Console.clear_formating()
    Console.newline()


@db_session
def new_category(conf):
    category_text = text_input("Category")

    Console.newline()
    Console.newline()

    try:
        new_category = database.Category(name=category_text)
        conf._db.add(new_category)
        row_update = conf._db.commit()

    except Exception as ex:
        Console.write(RED)
        Console.writeline("Update error. Category was not saved.")
        Console.clear_formating()
        return

    Console.write(GREEN)
    Console.write(f"New Category {category_text} saved!")
    Console.clear_formating()
    Console.newline()


@db_session
def _new_incident(conf):
    subject_text = text_input("Project/Incident")

    parse_incident_id = regex_factory(
        (
            r"^.*(?P<type>WO|INC|CRQ|FY\d\d)-?"
            r"0*(?P<number>[0-9]\d+) "
            r"(?P<description>.+)$"
        )
    )

    incident_id = parse_incident_id(subject_text) or {}

    Console.newline()

    prj_type = default_text_input("Type", incident_id.get("type", "WO"))
    prj_num = number_input("Number", incident_id.get("number", ""))
    prj_desc = default_text_input("Desc", incident_id.get("description", ""))

    Console.newline()

    try:
        project = database.Project(type=prj_type, number=prj_num, name=prj_desc)
        conf._db.add(project)
        conf._db.commit()
    except Exception as ex:
        Console.write(RED)
        Console.writeline("Update error. Incicent was not saved.")
        Console.clear_formating()
        return

    Console.write(GREEN)
    Console.write("New incident saved!")
    Console.clear_formating()
    Console.newline()


@db_session
def _time_entry(conf):
    """Creates a new entry. When doing so it ends the current entry
    (if there is one) and records its duration.

    When the new entry is created, the entries timer is displayed.

    """
    # fetch working data
    categories = conf._db.query(database.Category).all()
    projects = conf._db.query(database.Project).all()

    query = conf._db.query(database.Entry)

    query = query.order_by(desc(database.Entry.timestamp))
    last_entry = query.first()

    # get inputs
    category = _option_input("Category", categories)
    project = _option_input("Project", projects)

    note = text_input("Notes")
    timestamp = _date_input("Date", conf.dt)
    #  duration = duration_input("Duration")

    # calculate duration of last entry
    #  last_entry_id, last_entry_dur = conf.dt.get_duration(last_entry, timestamp)

    entry = database.Entry(
        category=category,
        project=project,
        note=note,
        timestamp=timestamp,
        duration=duration,
    )

    conf._db.add(entry)
    conf._db.commit()

    if last_entry:
        # calculate duration
        # update last entry
        pass

    show_timer(timestamp, conf.dt)


def time_entry(conf):
    """Creates a new entry. When doing so it ends the current entry
    (if there is one) and records its duration.

    When the new entry is created, the entries timer is displayed.

    """
    # fetch working data
    with Database(conf.sqlite, conf.dt) as db:

        categories = db.fetch_category()
        incidents = db.fetch_incidents()
        last_entry = db.fetch_last_entry()

    # get inputs
    category = _option_input("Category", categories)
    incident = _option_input("Incident", incidents)

    notes = text_input("Notes")
    timestamp = date_input("Date", conf.dt)
    duration = duration_input("Duration")

    # calculate duration of last entry
    last_entry_id, last_entry_dur = conf.dt.get_duration(last_entry, timestamp)

    with Database(conf.sqlite, conf.dt) as db:
        db.record_entry(category, incident, notes, timestamp, duration)
        if last_entry_id:
            db.update_entry_duration(last_entry_id, last_entry_dur)

    show_timer(timestamp, conf.dt)


def show_current_entry(conf):
    """Show the last or current entry and display it's timer.
    If there is no open entries, it shows nothing.

    """
    with Database(conf.sqlite, conf.dt) as db:
        last_entry = db.fetch_last_entry()

    if not last_entry or last_entry.minutes != 0:
        Console.write(RED)
        Console.writeline(" There are no active entries!")
        Console.clear_formating()
        return

    Console.write(str(last_entry))
    Console.newline()

    start_time = conf.dt.parse_date_input(last_entry.start_time)
    show_timer(start_time, conf.dt)


def edit_entry(conf):
    if not conf.cli_curr_obj:
        print("We don't have the selection feature built yet!")
        return

    with Database(conf.sqlite, conf.dt) as db:
        last_entry = db.fetch_last_entry()
        categories = db.fetch_category()
        incidents = db.fetch_incidents()

    # get inputs
    entry_id = last_entry.entry_id
    category = option_input("Category", categories, default=last_entry.category)
    incident = option_input("Incident", incidents, default=last_entry.incident)

    notes = text_input("Notes", default=last_entry.note)
    timestamp = date_input("Date", conf.dt, default=last_entry.start_time)

    do_update = bool_input("Proceed with update?")

    if not do_update:
        print("canceling update")
        return

    with Database(conf.sqlite, conf.dt) as db:
        db.update_entry(entry_id, category, incident, notes, timestamp)


def stop_entry(conf):
    """Ends the current entry (if there is one) and records its
    duration.

    """
    # fetch working data
    with Database(conf.sqlite, conf.dt) as db:
        last_entry = db.fetch_last_entry()

        if not last_entry:
            Console.write(RED)
            Console.writeline(" There are no entries!")
            return

        Console.writeline(str(last_entry))

        if last_entry.minutes > 0:
            Console.write(RED)
            Console.writeline(" Last entry already has a duration.")
            Console.clear_formating()
            return

        timestamp = date_input("End Date", conf.dt)
        duration = duration_input("Duration")

        # calculate duration of last entry
        last_entry_id, last_entry_dur = conf.dt.get_duration(last_entry, timestamp)

        # if we have a duration, then we override the calculated value
        if last_entry_id and duration:
            last_entry_dur = duration

        if last_entry_id:
            db.update_entry_duration(last_entry_id, last_entry_dur)


def summary(conf):
    """Displays a summary of the current days total time.

    """
    from tabulate import tabulate

    with Database(conf.sqlite, conf.dt) as db:
        row = db.fetch_todays_entry_summary()

    Console.write(BG_WHITE)
    Console.write(BLACK)
    Console.newline()

    table = tabulate(
        [row], headers=("DATE", "START", "END", "DURATION"), numalign="decimal"
    )
    Console.write(add_table_border(table))
    Console.newline()
    Console.clear_formating()


def weekly_summary(conf):
    """Displays a summary of times for a given range of days.

    """
    from tabulate import tabulate

    start_date = date_input("Start Date", conf.dt)
    end_date = date_input("End Date", conf.dt)

    with Database(conf.sqlite, conf.dt) as db:
        rows = db.fetch_weekly_totals(
            start_date.format("YYYY-MM-DD"), end_date.format("YYYY-MM-DD")
        )

    Console.write(BG_WHITE)
    Console.write(BLACK)

    table = tabulate(
        rows,
        numalign="decimal",
        headers=("INCIDENT", "TOTALS", "SUN", "MON", "TUE", "WED", "THU", "FRI", "SAT"),
    )

    Console.write(add_table_border(table))
    Console.newline()
    Console.clear_formating()


def view(conf):
    from tabulate import tabulate

    today = conf.dt.current_timestamp().format("YYYY-MM-DD")
    with Database(conf.sqlite, conf.dt) as db:
        rows = db.fetch_entry_view(today)

    Console.write(BG_WHITE)
    Console.write(BLACK)
    table = tabulate(
        rows,
        headers=("TIMESTAMP", "MIN", "CATEGORY", "INCIDENT", "NOTES"),
        numalign="decimal",
    )
    Console.write(add_table_border(table))
    Console.newline()
    Console.clear_formating()



from datetime import datetime
import pendulum
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import Text
from sqlalchemy import create_engine
from sqlalchemy import desc
from sqlalchemy import func
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.orm import sessionmaker
from . import time


Base = declarative_base()


def init(dbfile):
    engine = create_engine(f"sqlite:///{dbfile}")
    Base.metadata.create_all(bind=engine)
    return sessionmaker(bind=engine)


def create_default_config(Session, state):
    session = Session()
    Base.meta.create_all(session)
    session.commit()


class BaseTable:  # pylint: disable=R0903
    id = Column(Integer, primary_key=True, autoincrement=True)
    active = Column(Boolean(), default=True)
    created = Column(DateTime(timezone=False), default=pendulum.now)
    modified = Column(
        DateTime(timezone=False), default=pendulum.now, onupdate=pendulum.now
    )

    @classmethod
    def get_all(cls, db):
        return db.query(cls).all()

    @classmethod
    def get_active(cls, db):
        return db.query(cls).filter(cls.active == True).all()  # noqa


FMTS = {
    "INC": "{}{:0>8}",
    "WO": "{}{:0>9}",
    "FY16": "{}-{:0>6}",
    "FY17": "{}-{:0>6}",
    "FY18": "{}-{:0>6}",
    "FY19": "{}-{:0>6}",
    "FY20": "{}-{:0>6}",
    "FY21": "{}-{:0>6}",
    "FY22": "{}-{:0>6}",
    "": "{}-{:0>8}",
}


class Project(Base, BaseTable):
    __tablename__ = "project"

    type = Column(String(25), nullable=False)
    number = Column(Integer, nullable=False, unique=True)
    name = Column(String(100), nullable=False)
    persistant = Column(Boolean(), default=False)

    @property
    def identifier(self):
        return FMTS.get(self.type, FMTS[""]).format(self.type, self.number)


class Category(Base, BaseTable):
    __tablename__ = "category"
    name = Column(String(100), nullable=False, unique=True)


class Entry(Base, BaseTable):
    __tablename__ = "entry"
    category_id = Column(Integer, ForeignKey(Category.id))
    category = relationship(Category, lazy="immediate")
    project_id = Column(Integer, ForeignKey(Project.id))
    project = relationship(Project)
    note = Column(Text, default="")
    start_time = Column(DateTime(timezone=False), unique=True)
    end_time = Column(DateTime(timezone=False), unique=True)
    duration = Column(Integer)

    @property
    def is_completed(self):
        return self.end_time and self.end_time

    def start(self):
        self.start_time = datetime.now()

    def end(self):
        if not self.start_time:
            raise Exception("You cannot end an entry without a start time.")
        self.end_time = datetime.now()
        self.duration = self.calculate_duration()

    def calculate_duration(self):
        if not self.end_time or not self.end_time:
            raise Exception("You cannot end an entry without a start time.")

        elapsed_time = self.end_time - self.start_time
        return int(elapsed_time.total_seconds() / 60)

    @classmethod
    def get_day_start_entry(cls):
        entry = cls(note="start of day")
        entry.start_time = end_time = time.get_start_of_day()
        entry.end_time = end_time
        entry.duration = 0
        return entry

    @classmethod
    def get_last(cls, db, given_datetime):
        day = given_datetime.replace(hour=0, minute=0, second=0, microsecond=0)
        query = db.query(cls).filter(cls.start_time >= day)
        query = query.order_by(desc(cls.start_time))
        return query.first()

    @classmethod
    def current_week(cls, db):
        bow = time.get_start_of_week()
        print(f"start of week: {bow}")
        query = db.query(cls).filter(cls.start_time >= bow)
        query = query.order_by(cls.start_time)
        return query.all()

    @classmethod
    def day_summary(cls, db):
        today = time.get_start_of_day()
        query = db.query(func.min(cls.start_time), func.min(cls.start_time), func.max(cls.end_time), func.sum(cls.duration))
        query = query.filter(cls.start_time >= today)
        return query.all()

    @classmethod
    def week_summary(cls, db):
        today = time.get_start_of_week()
        query = db.query(Category.name, func.count(cls.id), func.sum(cls.duration))
        query = query.filter(cls.start_time >= today)
        return query.group_by(Category.name).all()



import codecs
import os
import re
import sqlite3
from sqlite3 import IntegrityError

from ttmr.domain import ViewEntry
from ttmr.util import DB_TIMESTAMP
from ttmr.util import find_data_file
from ttmr.util import replace_bad_char

codecs.register_error("replace", replace_bad_char)


def fetch_sql(sqlfile):
    with open(find_data_file(sqlfile), "r") as fh_:
        return fh_.read()


FMTS = {
    "INC": "{}{:0>8}",
    "WO": "{}{:0>9}",
    "FY16": "{}-{:0>6}",
    "FY17": "{}-{:0>6}",
    "FY18": "{}-{:0>6}",
    "FY19": "{}-{:0>6}",
    "FY20": "{}-{:0>6}",
}


def fmt_inc(typ, number):
    """
    INC00000000
    WO000000000
    FY18-000000
    """
    return FMTS[typ].format(typ, number)


class Database(object):
    def __init__(self, dbfile, dt):
        self.dbfile = dbfile
        self.dt = dt

    def __enter__(self):
        self.conn = sqlite3.connect(self.dbfile)
        self.csr = self.conn.cursor()
        return self

    @property
    def version(self):
        return sqlite3.sqlite_version

    def initialize(self):
        sql = fetch_sql("INITIALIZE.sql")
        self.csr.executescript(sql)

    def select(self, sql, parms=None):
        try:
            if parms:
                self.csr.execute(sql, parms)
            else:
                self.csr.execute(sql)
            return self.csr.fetchall()
        except IntegrityError as ex:
            raise ex

    def fetch_last_record(self, table):
        sql = "SELECT * FROM {} ORDER BY ID DESC LIMIT 1;"
        self.csr.execute(sql.format(table))
        return self.csr.fetchone()

    def execute(self, sql, parms, raise_errors=False):
        try:
            self.csr.execute(sql, parms)
        except IntegrityError as ex:
            print(parms["number"], ex)
            if raise_errors:
                raise ex

    def commit(self):
        self.conn.commit()

    def __exit__(self, *args):
        self.conn.close()

    def record_category(self, *data):
        self.execute(self.new_category_sql, data)
        self.commit()

    def fetch_category(self):
        return self.select(u"SELECT ID, NAME FROM CATEGORY WHERE ACTIVE = 1;")

    def record_incident(self, *data):
        # input data
        typ, num, name = data
        timestamp = self.dt.current_timestamp().format(DB_TIMESTAMP)
        parms = {
            "typ": typ,
            "number": num,
            "name": name,
            "active": 1,
            "persist": 0,
            "created": timestamp,
            "modified": timestamp,
        }

        self.execute(fetch_sql("INSERT-INCIDENT.sql"), parms)
        self.commit()
        return self.csr.rowcount

    def fetch_incidents(self):
        return self.select(u"SELECT ID, NAME FROM VIEW_INCIDENTS;")

    def deactivate_incident(self, inc_id):
        parms = {"id": inc_id}
        self.execute(
            ("UPDATE INCIDENT SET ACTIVE = 0 " "WHERE ID = :id AND PERSISTANT <> 1;"),
            parms,
        )
        self.commit()
        return self.csr.rowcount

    def fetch_last_entry(self):
        sql = u"SELECT * FROM VIEW_ENTRY ORDER BY ENTRY_ID DESC LIMIT 1;"
        self.csr.execute(sql)
        row = self.csr.fetchone()
        if not row:
            return row
        return ViewEntry(*row)

    def record_category(self, *data):
        self.execute(fetch_sql("INSERT-CATEGORY.sql"), data, raise_errors=True)
        self.commit()

    def record_entry(self, *data):
        category, incident, notes, timestamp, duration = data
        created = self.dt.current_timestamp().format(DB_TIMESTAMP)
        parms = {
            "category": category,
            "incident": incident,
            "note": notes,
            "timestamp": timestamp.format(DB_TIMESTAMP),
            "duration": duration,
            "created": created,
            "modified": created,
        }
        self.execute(fetch_sql("INSERT-ENTRY.sql"), parms, raise_errors=True)
        self.commit()

    def update_entry(self, *data):
        _id, category, incident, notes, timestamp = data
        modified = self.dt.current_timestamp().format(DB_TIMESTAMP)
        parms = {
            "id": _id,
            "category": category,
            "incident": incident,
            "note": notes,
            "timestamp": timestamp.format(DB_TIMESTAMP),
            "modified": modified,
        }
        self.execute(fetch_sql("UPDATE-ENTRY.sql"), parms, raise_errors=True)
        self.commit()

    def fetch_todays_entry_summary(self):
        sql = u"SELECT * FROM VIEW_CURRENT_TIME_SUMMARY"
        self.csr.execute(sql)
        return self.csr.fetchone()

    def fetch_weekly_totals(self, start_date, end_date):
        sql = fetch_sql("SELECT-WEEK-SUMMARY.sql")
        rows = self.select(sql, parms={"startdate": start_date, "enddate": end_date})
        return rows

    def fetch_entry_view(self, day):
        sql = fetch_sql("SELECT-ENTRY-VIEW.sql")
        rows = self.select(sql, parms={"day": day})
        return [ViewEntry(*r).to_view() for r in rows]

    def update_entry_duration(self, entry_id, duration):
        created = self.dt.current_timestamp().format(DB_TIMESTAMP)
        sql = fetch_sql("UPDATE-ENTRY-DURATION.sql")
        self.execute(
            sql, {"dur": duration, "id": entry_id, "mod": created}, raise_errors=True
        )
        self.commit()


def fetch_directory_incidences(conf):
    """Look for incidences from directory names.

    """
    reg = re.compile(
        (r"^(?P<type>WO|INC|CRQ)" r"0*(?P<number>[0-9]\d+) " r"?- ?(?P<name>.*)\w?$"),
        re.I,
    )

    if not os.path.exists(conf.project_dir):
        print("The project directory does not exist. Nothing to load.")
        return

    for entry in os.listdir(conf.project_dir):
        m = reg.match(entry)
        if m:
            g = m.groupdict()
            yield (g["type"], int(g["number"]), g["name"].strip())


def fetch_file_incidences(conf):
    """Look for incidences from a files content.

    """
    # get from list
    reg = re.compile((r"^(?P<type>WO|INC|CRQ|FY\d\d)" r"0*(?P<number>[0-9]\d+)$"), re.I)

    remedyfile = os.path.join(conf.project_dir, "remedy-tickets.txt")

    if not os.path.exists(remedyfile):
        print("Remedy file does not exist!")
        return

    with open(remedyfile) as fh_:
        lines = fh_.readlines()

    parse_incident_id = regex_factory(
        (r"^(?P<type>WO|INC|CRQ)" r"0*(?P<number>[0-9]\d+)$")
    )

    ilines = iter(lines)
    for line in ilines:
        g = parse_incident_id(line)
        if g:
            next(ilines)  # skip a line
            g["name"] = next(ilines).strip()
            yield (g["type"], int(g["number"]), g["name"].strip())


def load_incidences(conf):

    import pyperclip

    remedy_regex = "^(?P<type>(INC|WO))0*(?P<id>[1-9]\d+)[\r\n]+Mark Gemmill[\r\n]+^(?P<name>[^\r\n]+)[\r\n]?$"
    outlook_regex = "(?P<type>(INC|WO))0*(?P<id>[1-9]\d+).*Description: (?P<name>.+)$"
    er_regex = "(?P<type>FY\d\d)-0*(?P<id>[1-9]\d+)[^\r\n]*[\r\n]+^(?P<name>[^\r\n]+)[\r\n]+^Mark Gemmill"
    jira_regex = "(?P<type>(INC|WO|FY\d\d))[^\d]*0*(?P<id>[1-9]\d+) +(?P<name>[^\r\n]+)[\r\n]+^\S+"

    re_flags = re.M

    capture_list = [
        re.compile(remedy_regex, re_flags),
        re.compile(outlook_regex, re_flags),
        re.compile(er_regex, re_flags),
        re.compile(jira_regex, re_flags),
    ]

    def parse_from(regex, pasted):
        results = []
        for result in regex.finditer(pasted):
            d = result.groupdict()
            results.append((d["type"].upper(), d["id"], d["name"]))
        return results

    def parse_pasted():
        content = pyperclip.paste()
        for rx in capture_list:
            for result in parse_from(rx, content):
                yield result

    with Database(conf.sqlite, conf.dt) as db:
        print("loading....")

        for typ, num, nme in parse_pasted():
            print("  {} - {}".format(fmt_inc(typ, num), nme))
            try:
                db.record_incident(typ, num, nme)
            except Exception as ex:
                print(ex)



import os
from pathlib import Path
from types import SimpleNamespace

import pendulum
import tomlkit
from appdirs import AppDirs

DEFAULT_CONFIG = """
[ttmr]
project_dir=
timezone=US/Pacific
"""


def init_file(path):
    if not os.path.exists(path):
        with open(path, "w") as fh_:
            fh_.write("")


def assert_dir(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)


def get_appdir():
    appdirs = AppDirs("ttmr", "mgemmill")
    user_data_dir = appdirs.user_data_dir
    assert_dir(user_data_dir)
    return user_data_dir


#  class Config(object):
#      def __str__(self):
#          output = []
#          for k, v in self.__dict__.items():
#              if not k.startswith("_"):
#                  output.append("{}={}".format(k, v))
#          return "\n".join(output)
#
#      def load_cli_args(self, args):
#          for key, val in args.items():
#              normalize_key = key.strip("-").replace("-", "_").lower()
#              setattr(self, "cli_" + normalize_key, val)


def get_config():
    cfg = SimpleNamespace()

    cfg.appdirs = appdirs = AppDirs("ttmr", "mgemmill")
    cfg.user_data_dir = app_dir = appdirs.user_data_dir

    assert_dir(app_dir)

    cfg.config_path = Path(app_dir, "ttmr.cfg").resolve()
    cfg.data = data = tomlkit.parse(cfg.config_path.read_text())
    cfg.db_path = Path(app_dir, data["ttmr"]["db"])
    cfg.timezone = timezone = data["ttmr"].get("timezone", "America/Vancouver")

    today = pendulum.today(timezone)
    hour, minute = [int(s) for s in data["ttmr"].get("start_time", "07:00").split(":")]

    cfg.day_start_time = pendulum.datetime(
        year=today.year, month=today.month, day=today.day, hour=hour, minute=minute
    )

    return cfg


#  def config():
#
#      c = Config()
#
#      c.appdir = appdir = get_appdir()
#      c.configfile = os.path.join(appdir, "ttmr.ini")
#
#      config_path = Path(appdir, 'ttmr.cfg').resolve()
#
#
#      c.sqlite = os.path.join(appdir, "ttmr.db3")
#
#      if not os.path.exists(c.configfile):
#          with open(c.configfile, "w") as fh_:
#              fh_.write(DEFAULT_CONFIG)
#
#      parser = SafeConfigParser()
#      parser.read(c.configfile)
#
#      c.project_dir = parser.get("ttmr", "project_dir")
#      c.timezone = parser.get("ttmr", "timezone")
#
#      c.dt = dt = DT(c.timezone)
#
#      if not os.path.exists(c.sqlite):
#          with Database(c.sqlite, dt) as db:
#              db.initialize()
#
#      c._dbfile = Path(appdir, 'ttmr.db')
#
#      c._cfg = tomlkit.parse(config_path.read_text())
#
#      return c



from datetime import datetime
from ..gui import date_input
from ..gui import duration_input
from ..gui import option_input
from ..gui import show_timer
from ..ui import text_input
from ..database import Category
from ..database import Entry
from ..database import Project
from .util import db_session
from .util import new_session


def time_entry(conf, ui, end_entry=False):
    """Creates a new entry. When doing so it ends the current entry
    (if there is one) and records its duration.

    When the new entry is created, the entries timer is displayed.

    The timer requires that there is a Day Start Entry.
    Day Start will have a start_time and end_time that are the same.

    There are 2 ways to create entries:

    1. as a ttmr start command:
        - this will open a new entry with a start time that is set
          automatically based on the end_time of the previous entry.
        - if there is no previous entry for the day, then it will use
          the default day's start time.
        - the window remains open with a time counter that explains
          this is the count on the current timer.
        - when you close the window, the end time will be set.

    2. as a ttmr end command:
        - this will open a new entry with a start that is set
          automatically based on the end_time of the previous entry.
        - if there is no previous entry for the day, then it will use
          the default day's start time.
        - this will also set the end time.
        - the window remains open with time counter

    """
    # fetch working data
    categories = Category.get_all(conf.db)
    projects = Project.get_all(conf.db)

    end_time = None
    duration = 0

    current_timestamp = datetime.now()

    if end_entry:
        last_entry = Entry.get_last(conf.db, current_timestamp)

        if not last_entry:
            last_entry = Entry.get_day_start_entry()

        if last_entry.end_time is not None:
            current_timestamp = last_entry.end_time

    # get inputs
    category, project, note, start_time = ui.fetch_entry_inputs(
        categories, projects, current_timestamp
    )

    # if this is an end entry...
    if end_entry:
        # TODO: it is a bit of a problem that we might be overriding the duration...
        # TODO: and that the duration is a captured value
        end_time, duration = ui.fetch_end_entry_inputs(start_time, current_timestamp)

    # create entry
    entry = Entry(
        category=category,
        project=project,
        note=note,
        start_time=start_time,
        end_time=end_time,
        duration=duration,
    )

    conf.db.add(entry)
    conf.db.commit()

    return entry.id


@db_session
def start_time_entry(cfg):
    entry_id = time_entry(cfg, EntryUI, end_entry=False)
    EntryUI.post_entry_timer(cfg, "Current Project Timer:", datetime.now(), entry_id)


@db_session
def stop_time_entry(cfg):
    entry_id = time_entry(cfg, EntryUI, end_entry=True)
    EntryUI.post_entry_timer(cfg, "Time since last completed entry::", datetime.now())


class EntryUI:
    @staticmethod
    def fetch_entry_inputs(categories, projects, last_end_time):
        category = option_input("Category", categories)
        project = option_input("Project", projects)
        note = text_input("Notes")
        start_time = date_input("Start Date", default=last_end_time)
        return category, project, note, start_time

    @staticmethod
    def fetch_end_entry_inputs(start_time, current_timestamp):
        end_time = date_input("End Date", default=current_timestamp)
        elapsed_time = 0
        elapsed_time = current_timestamp - start_time
        duration = duration_input(
            "Duration", str(int(elapsed_time.total_seconds() / 60))
        )
        return end_time, duration

    @staticmethod
    def post_entry_timer(cfg, msg, current_timestamp, current_entry_id=None):
        print()

        if current_entry_id:
            print(msg)
            try:
                show_timer(current_timestamp)
            except KeyboardInterrupt:
                print("Closing current session...")
                with new_session(cfg.db_path) as db:
                    current_entry = (
                        db.query(Entry)
                        .filter(Entry.id == current_entry_id)
                        .one_or_none()
                    )
                    current_entry.end()
                    db.commit()
        else:
            print(msg)
            show_timer(current_timestamp)



import sys
import threading
from contextlib import contextmanager
from datetime import datetime
from functools import wraps

from sqlalchemy import desc
from tabulate import tabulate

from ttmr.console import BG_WHITE
from ttmr.console import BLACK
from ttmr.console import GREEN
from ttmr.console import RED
from ttmr.console import Console
from ttmr.db import Database
from ttmr.db import load_incidences
from ttmr.db import regex_factory
from ttmr.gui import date_input
from ttmr.gui import duration_input
from ttmr.gui import number_input
from ttmr.gui import option_input
from ttmr.gui import show_timer
from ttmr.ui import default_text_input
from ttmr.ui import text_input
from ttmr.util import add_table_border

from . import database
from .database import Category
from .database import Entry
from .database import Project
from .util import db_session
from .util import new_session


def display_config(conf):
    Console.write(str(conf))


def load(conf):
    load_incidences(conf)


@db_session
def new_category(conf):
    category_text = text_input("Category")

    Console.newline()
    Console.newline()

    try:
        new_category = Category(name=category_text)
        conf.db.add(new_category)
        row_update = conf.db.commit()

    except Exception as ex:
        Console.write(RED)
        Console.writeline("Update error. Category was not saved.")
        Console.clear_formating()
        return

    Console.write(GREEN)
    Console.write(f"New Category {category_text} saved!")
    Console.clear_formating()
    Console.newline()


@db_session
def list_categories(conf):

    categories = Category.get_all(conf.db)

    Console.write(BG_WHITE)
    Console.write(BLACK)

    table = tabulate(
        [(c.name, c.active, c.created, c.modified) for c in categories],
        headers=("CATEGORY", "ACTIVE", "CREATED", "MODIFED"),
        numalign="decimal",
    )
    Console.write(add_table_border(table))
    Console.newline()
    Console.clear_formating()


@db_session
def new_project(conf):
    subject_text = text_input("Project/Incident")

    parse_incident_id = regex_factory(
        (
            r"^.*(?P<type>WO|INC|CRQ|FY\d\d)-?"
            r"0*(?P<number>[0-9]\d+) "
            r"(?P<description>.+)$"
        )
    )

    incident_id = parse_incident_id(subject_text) or {}

    Console.newline()

    prj_type = default_text_input("Type", incident_id.get("type", "WO"))
    prj_num = number_input("Number", incident_id.get("number", ""))
    prj_desc = default_text_input("Desc", incident_id.get("description", ""))

    Console.newline()

    try:
        project = Project(type=prj_type, number=prj_num, name=prj_desc)
        conf.db.add(project)
        conf.db.commit()
    except Exception as ex:
        Console.write(RED)
        Console.writeline("Update error. Incicent was not saved.")
        Console.writeline(f"(Error: {ex})")
        Console.clear_formating()
        return

    Console.write(GREEN)
    Console.write("New incident saved!")
    Console.clear_formating()
    Console.newline()


@db_session
def list_projects(conf):
    projects = Project.get_all(conf.db)

    Console.write(BG_WHITE)
    Console.write(BLACK)

    table = tabulate(
        [(c.identifier, c.name, c.active, c.created, c.modified) for c in projects],
        headers=("IDENTIFIER", "NAME", "ACTIVE", "CREATED", "MODIFED"),
        numalign="decimal",
    )
    Console.write(add_table_border(table))
    Console.newline()
    Console.clear_formating()


@db_session
def start_day(conf):
    """
    Set the begging time of the day.

    """
    start_category = (
        conf.db.query(Category).filter(Category.name == "Day Start").one_or_none()
    )
    start_project = (
        conf.db.query(Project).filter(Project.name == "Day Start").one_or_none()
    )

    timestamp = datetime.now()

    entry = Entry(
        category=start_category,
        project=start_project,
        note="Day Start",
        timestamp=timestamp,
        duration=0,
    )

    conf.db.add(entry)
    conf.db.commit()

    print()

    show_timer(timestamp)


@db_session
def time_entry(conf, start_entry=False):
    """Creates a new entry. When doing so it ends the current entry
    (if there is one) and records its duration.

    When the new entry is created, the entries timer is displayed.

    The timer requires that there is a Day Start Entry.
    Day Start will have a start_time and end_time that are the same.

    There are 2 ways to create entries:

    1. as a ttmr start command:
        - this will open a new entry with a start time that is set
          automatically based on the end_time of the previous entry.
        - if there is no previous entry for the day, then it will use
          the default day's start time.
        - the window remains open with a time counter that explains
          this is the count on the current timer.
        - when you close the window, the end time will be set.

    2. as a ttmr end command:
        - this will open a new entry with a start that is set
          automatically based on the end_time of the previous entry.
        - if there is no previous entry for the day, then it will use
          the default day's start time.
        - this will also set the end time.
        - the window remains open with time counter

    """
    # fetch working data
    categories = Category.get_all(conf.db)
    projects = Project.get_all(conf.db)

    current_timestamp = datetime.now()

    last_entry = Entry.get_last(conf.db, today)

    if not last_entry:
        last_entry = Entry.get_day_start_entry()

    if last_entry.end_time is None:
        last_entry.end_time = current_timestamp

    # get inputs
    category = option_input("Category", categories)
    project = option_input("Project", projects)
    note = text_input("Notes")
    start_time = date_input("Start Date", default=last_entry.end_time)

    end_time = None
    duration = 0

    if not start_entry:
        end_time = date_input("End Date", default=current_timestamp)
        elapsed_time = 0
        elapsed_time = current_timestamp - start_time
        duration = duration_input(
            "Duration", str(int(elapsed_time.total_seconds() / 60))
        )

    # create entry
    entry = Entry(
        category=category,
        project=project,
        note=note,
        start_time=last_entry.end_time,
        end_time=current_end_time,
        duration=duration,
    )

    conf.db.add(entry)
    conf.db.commit()
    conf.db.close()

    current_entry_id = entry.id

    print()

    if start_entry:
        print("Current Entry Time:")
        try:
            show_timer(current_timestamp)
        except KeyboardInterrupt:
            print("Closing current session...")
            with new_session(cfg.db_path) as db:
                current_entry = (
                    db.query(Entry).filter(Entry.id == current_entry_id).one_or_none()
                )
                current_entry.end()
                db.commit()
    else:
        print("Time since last entry:")
        show_timer(current_timestamp)


def _list_entries(list_of_entries):
    entries = [
        (e.category.name, e.project.identifier, e.note, e.timestamp, e.duration)
        for e in list_of_entries
    ]

    total_minutes = int(sum([e[4] for e in entries]))

    entries.append(("", "", "", "", total_minutes))

    Console.write(BG_WHITE)
    Console.write(BLACK)

    table = tabulate(
        entries,
        headers=("CATEGORY", "PROJECT", "NOTE", "TIMESTAMP", "DURATION"),
        numalign="decimal",
    )
    Console.write(add_table_border(table))
    Console.newline()
    Console.clear_formating()


@db_session
def list_entries(conf):
    _list_entries(Entry.get_all(conf.db))


@db_session
def list_current_weeks_entries(conf):
    _list_entries(Entry.current_week(conf.db))



"""
 Usage:
     ttmr current
     ttmr start
     ttmr stop
     ttmr view
     ttmr summary (--day|--week)
     ttmr weekly
     ttmr new  OBJTYPE
     ttmr list OBJTYPE [--inactive]

 Options:
   -h --help     Show this screen.
   --version     Show version.

"""
import codecs

import colorama

import ttmr.config as config
from ttmr.console import GREEN
from ttmr.console import RED
from ttmr.console import Console
from ttmr.util import replace_bad_char
from . import commands

__author__ = "Mark Gemmill"
__email__ = "mark@markgemmill.com"
__version__ = "0.5.2-alpha"


colorama.init()

codecs.register_error("ttmr", replace_bad_char)


def main_wrapper(func):
    def _wrapper_():
        try:
            Console.init()
            Console.newline()
            Console.write(GREEN)
            Console.write(" Task Timer v{}".format(__version__))
            Console.newline()
            Console.write(" -----------------")
            Console.clear_formating()
            Console.newline()
            Console.cursor_up()

            func()

            Console.newline()

        except KeyboardInterrupt:
            Console.clear_formating()
            Console.newline()
            Console.newline()
            Console.write(RED)
            Console.writeline("*** Task has been canceled. ***")
            Console.clear_formating()
            Console.newline()

    return _wrapper_


@main_wrapper
def main():
    import docopt

    conf = config.get_config()
    args = docopt.docopt(__doc__, version="")
    conf.cli_args = args

    Console.newline()

    if args["stop"]:
        commands.stop_time_entry(conf)
    elif args["start"]:
        commands.start_time_entry(conf)
    elif args["current"]:
        #  cmd.show_current_entry(conf)
        pass
    elif args["summary"] and args["--day"]:
        commands.day_summary(conf)
    elif args["summary"] and args["--week"]:
        commands.weekly_summary(conf)
    elif args["weekly"]:
        #  cmd.weekly_summary(conf)
        commands.list_current_weeks_entries(conf)
    elif args["view"]:
        #  cmd.view(conf)
        pass
    #  elif args["load"]:
    #  cmd.load(conf)
    #  pass
    #  elif args["config"]:
    #  cmd.display_config(conf)
    #  pass
    #  elif args["edit"] and args["OBJTYPE"] == "ent":
    #  cmd.edit_entry(conf)
    #  pass
    elif args["new"] and args["OBJTYPE"] in ("cat", "category"):
        commands.new_category(conf)
    elif args["list"] and args["OBJTYPE"] in ("cat", "category"):
        commands.list_categories(conf)
    elif args["new"] and args["OBJTYPE"] in (
        "prj",
        "proj",
        "inc",
        "project",
        "incident",
    ):
        commands.new_project(conf)
    elif args["list"] and args["OBJTYPE"] in (
        "prj",
        "proj",
        "inc",
        "project",
        "incident",
    ):
        commands.list_projects(conf)
    elif args["list"] and args["OBJTYPE"] in ("ent", "entry"):
        commands.list_entries(conf)
    #  elif args["start"] and args["day"]:
    #      commands.start_day(conf)
    else:
        print("Unknown command: could not complete your request.")



from tabulate import tabulate
from ..console import BG_WHITE
from ..console import BLACK
from ..console import Console
from ..ui import default_text_input
from ..ui import text_input
from ..util import add_table_border
from ..database import Category
from ..database import Entry
from ..database import Project
from .util import db_session
from .. import gui


def display_config(conf):
    Console.write(str(conf))


def load(conf):
    load_incidences(conf)


@db_session
def list_categories(conf):

    categories = Category.get_all(conf.db)

    Console.write(BG_WHITE)
    Console.write(BLACK)

    table = tabulate(
        [(c.name, c.active, c.created, c.modified) for c in categories],
        headers=("CATEGORY", "ACTIVE", "CREATED", "MODIFED"),
        numalign="decimal",
    )
    Console.write(add_table_border(table))
    Console.newline()
    Console.clear_formating()


@db_session
def list_projects(conf):
    projects = Project.get_all(conf.db)

    Console.write(BG_WHITE)
    Console.write(BLACK)

    table = tabulate(
        [(c.identifier, c.name, c.active, c.created, c.modified) for c in projects],
        headers=("IDENTIFIER", "NAME", "ACTIVE", "CREATED", "MODIFED"),
        numalign="decimal",
    )
    Console.write(add_table_border(table))
    Console.newline()
    Console.clear_formating()


def _list_entries(list_of_entries):
    entries = [
        (
            e.category.name,
            e.project.identifier,
            e.note,
            e.start_time,
            e.end_time,
            e.duration,
        )
        for e in list_of_entries
    ]

    total_minutes = int(sum([e[5] for e in entries]))

    entries.append(("", "", "", "", "", total_minutes))

    Console.write(BG_WHITE)
    Console.write(BLACK)

    table = tabulate(
        entries,
        headers=("CATEGORY", "PROJECT", "NOTE", "START", "END", "DURATION"),
        numalign="decimal",
    )
    Console.write(add_table_border(table))
    Console.newline()
    Console.clear_formating()


@db_session
def list_entries(conf):
    _list_entries(Entry.get_all(conf.db))


@db_session
def list_current_weeks_entries(conf):
    _list_entries(Entry.current_week(conf.db))


@db_session
def day_summary(conf):
    """
    Displays a summary of the current days total time.

    """
    rows = Entry.day_summary(conf.db)
    gui.write_table(rows, ("DAY", "START", "END", "MINUTES"), ("{:%Y-%m-%d}", "{:%H:%M}", "{:%H:%M}", None))


@db_session
def weekly_summary(conf):
    """Displays a summary of times for a given range of days.

    """
    rows = Entry.week_summary(conf.db)
    gui.write_table(rows, ("CATEGORY", "ENTRIES", "MINUTES"))



from datetime import datetime
from datetime import timedelta
import arrow
from dateutil import tz


DEFAULT_TMSTP_FMT = "YYYY-MM-DD HH:mm"

DB_TIMESTAMP = "YYYY-MM-DD HH:mm:ss"

DATE_ENTRY_FORMATS = [DEFAULT_TMSTP_FMT, "YYYYMMDD HH:mm", "MM/DD/YYYY HH:mm"]


def get_start_of_week():
    """
    This will return the preceding Sunday at 12:00:00 am.

    """
    today = datetime.now()
    beginning_of_week = today - timedelta(days=(today.weekday() + 1))
    beginning_of_week = beginning_of_week.replace(
        hour=0, minute=0, second=0, microsecond=0
    )
    return beginning_of_week


def get_start_of_day(default_start_time=7):
    """
    This returns the current days default start time or the current time -
    which ever is earlier.

    """
    now = datetime.now()

    current_hour = now.hour if now.hour < 7 else 7
    current_minute = now.minute if now.hour < 7 else 0

    return now.replace(
        hour=current_hour, minute=current_minute, second=0, microsecond=0
    )


class DT(object):
    def __init__(self, timezone):
        self.tz = tz.gettz(timezone)

    def parse_date_input(self, date_input):
        #  return arrow.get(date_input, DATE_ENTRY_FORMATS, tzinfo=self.tz)
        if isinstance(date_input, datetime):
            return date_input
        return datetime.strptime(date_input, DATE_ENTRY_FORMATS)

    def current_timestamp(self):
        return arrow.now(tz=self.tz)

    def to_str(self, dt):
        return dt.strftime("%Y-%m-%d %H:%M")

    def calc_duration(self, end_time, start_time):
        # return minutes
        start_time = start_time.datetime
        end_time = end_time.datetime
        duration = end_time - start_time
        return int(duration.total_seconds() / 60.0)

    def get_duration(self, entry, now):
        if not entry or entry.minutes > 0:
            return None, None
        start_time = self.parse_date_input(entry.start_time)
        duration = self.calc_duration(now, start_time)
        return entry.entry_id, duration


class Timer(object):
    def __init__(self, start_time):  # , timezone):
        #  self.timezone = timezone
        if isinstance(start_time, datetime):
            self.start_time = start_time
        elif isinstance(start_time, str):
            self.start_time = datetime.strptime(start_time, "%Y-%m-%d %H:%M:%S")
        else:
            self.start_time = start_time

    def get_lapsed_seconds(self):
        end_time = datetime.now()
        duration = end_time - self.start_time
        return duration.total_seconds()

    def get_formatted_lapsed_time(self):
        total_seconds = self.get_lapsed_seconds()
        seconds = int(total_seconds) % 60
        total_minutes = int(total_seconds) / 60
        minutes = int(total_minutes % 60)
        hours = int(int(total_minutes) / 60)
        return "{:0>2}:{:0>2}:{:0>2}".format(hours, minutes, seconds)



import os
import re
import sys
from datetime import datetime

import arrow
import pendulum
from dateutil import tz

DEFAULT_TMSTP_FMT = "YYYY-MM-DD HH:mm"

DB_TIMESTAMP = "YYYY-MM-DD HH:mm:ss"

DATE_ENTRY_FORMATS = [DEFAULT_TMSTP_FMT, "YYYYMMDD HH:mm", "MM/DD/YYYY HH:mm"]


class DT(object):
    def __init__(self, timezone):
        self.tz = tz.gettz(timezone)

    def parse_date_input(self, date_input):
        #  return arrow.get(date_input, DATE_ENTRY_FORMATS, tzinfo=self.tz)
        if isinstance(date_input, datetime):
            return date_input
        return datetime.strptime(date_input, DATE_ENTRY_FORMATS)

    def current_timestamp(self):
        return arrow.now(tz=self.tz)

    def to_str(self, dt):
        return dt.strftime("%Y-%m-%d %H:%M")

    def calc_duration(self, end_time, start_time):
        # return minutes
        start_time = start_time.datetime
        end_time = end_time.datetime
        duration = end_time - start_time
        return int(duration.total_seconds() / 60.0)

    def get_duration(self, entry, now):
        if not entry or entry.minutes > 0:
            return None, None
        start_time = self.parse_date_input(entry.start_time)
        duration = self.calc_duration(now, start_time)
        return entry.entry_id, duration


def replace_bad_char(ex):
    return (u"?", ex.start)


class Timer(object):
    def __init__(self, start_time):  # , timezone):
        #  self.timezone = timezone
        if isinstance(start_time, datetime):
            self.start_time = start_time
        elif isinstance(start_time, str):
            self.start_time = datetime.strptime(start_time, "%Y-%m-%d %H:%M:%S")
        else:
            self.start_time = start_time

    def get_lapsed_seconds(self):
        end_time = pendulum.now()
        duration = end_time - self.start_time
        return duration.total_seconds()

    def get_formatted_lapsed_time(self):
        total_seconds = self.get_lapsed_seconds()
        seconds = int(total_seconds) % 60
        total_minutes = int(total_seconds) / 60
        minutes = int(total_minutes % 60)
        hours = int(int(total_minutes) / 60)
        return "{:0>2}:{:0>2}:{:0>2}".format(hours, minutes, seconds)


def find_data_file(filename):
    if getattr(sys, "frozen", False):
        # The application is frozen
        datadir = os.path.dirname(sys.executable)
    else:
        # The application is not frozen
        # Change this bit to match where you store your data files:
        datadir = os.path.dirname(__file__)
    return os.path.join(datadir, "sql", filename)


def add_table_border(text, border=1):
    lines = [(" " * border) + line for line in text.split("\n")]
    length = max([len(a) for a in lines])

    return "\n".join([l.ljust(length) for l in lines])


def regex_factory(regex):
    """Convenience wrapper around regex with named capture groups.
    """
    reg = re.compile(regex, re.I)

    def _match(text):
        m = reg.match(text)
        if m:
            return m.groupdict()

    return _match



from collections import namedtuple

_entry = namedtuple(
    "Entry",
    (
        "entry_id category_id category "
        "incident_id incident note "
        "start_time minutes hours"
    ),
)


class ViewEntry(_entry):

    __slots__ = ()

    def to_view(self):
        return (self.start_time, self.minutes, self.category, self.incident, self.note)

    def __str__(self):
        return (
            "   Category: {0.category}\n"
            "   Incident: {0.incident}\n"
            "       Note: {0.note}\n"
            " Start time: {0.start_time}\n"
            "   Duration: {0.minutes}\n"
        ).format(self)


def calc_hours(minutes):
    hours = minutes / 60.0
    return round(hours, 2)
